// See the end of this file for copyright and license terms.

use std::fmt;

/// Any primitive data type a [`Val`] can hold.
pub enum Type {
    Bool,
    Int,
    String,
}

/// The (immutable) result of evaluating an [`Expr`](super::expr::Expr).
pub trait Val : fmt::Display {
    fn clone(&self) -> Box<dyn Val>;

    fn typ(&self) -> Type;

    fn bool(&self) -> Result<&bool, TypeError> {
        Err(TypeError::new(self.typ(), Type::Bool))
    }

    fn int(&self) -> Result<&i64, TypeError> {
        Err(TypeError::new(self.typ(), Type::Int))
    }

    fn string(&self) -> Result<&str, TypeError> {
        Err(TypeError::new(self.typ(), Type::String))
    }

    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        if let Ok(s) = self.string() {
            write!(f, "[{} {}]", self.typ(), s)
        } else {
            write!(f, "[{} ??]", self.typ())
        }
    }
}

/// The actual data structure behind [`Val`].
struct Data<T: Clone> {
    data: T,
}

/// Construct a new bool value
pub fn bool(val: bool) -> Box<dyn Val> {
    Box::new(Data::new(val))
}

/// Construct a new int value
pub fn int(val: i64) -> Box<dyn Val> {
    Box::new(Data::new(val))
}

/// Construct a new string value
pub fn string(val: &str) -> Box<dyn Val> {
    Box::new(Data::new(String::from(val)))
}

impl<T: Clone> Data<T> {
    pub fn new(data: T) -> Data<T> {
        Data {
            data
        }
    }
}

impl Val for Data<bool> {
    fn clone(&self) -> Box<dyn Val> {
        Box::new(Data {
            data: self.data
        })
    }

    fn typ(&self) -> Type {
        Type::Bool
    }

    fn bool(&self) -> Result<&bool, TypeError> {
        Ok(&self.data)
    }
}

impl fmt::Display for Data<bool> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "[bool {}]", self.data)
    }
}

impl Val for Data<i64> {
    fn clone(&self) -> Box<dyn Val> {
        Box::new(Data {
            data: self.data
        })
    }

    fn typ(&self) -> Type {
        Type::Int
    }

    fn int(&self) -> Result<&i64, TypeError> {
        Ok(&self.data)
    }
}

impl fmt::Display for Data<i64> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "[int {}]", self.data)
    }
}

impl Val for Data<String> {
    fn clone(&self) -> Box<dyn Val> {
        Box::new(Data {
            data: self.data.clone()
        })
    }

    fn typ(&self) -> Type {
        Type::String
    }

    fn string(&self) -> Result<&str, TypeError> {
        Ok(&self.data)
    }
}

impl fmt::Display for Data<String> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "[string \"{}\"", self.data)
    }
}

pub struct TypeError {
    pub msg: String,
    pub actual: Type,
    pub expected: Type,
}

impl TypeError {
    pub fn new(actual: Type, expected: Type) -> TypeError {
        TypeError {
            msg: format!("Cannot convert from {} to {}", actual, expected),
            actual,
            expected,
        }
    }
}

impl Type {
    pub fn name(&self) -> &'static str {
        match self {
            Type::Bool => "bool",
            Type::Int => "int",
            Type::String => "string",
        }
    }
}

impl fmt::Display for Type {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "\"{}\"", self.name())
    }
}

// Copyright (C) 2021 Fefie <owo@fef.moe>
// This file is part of gayconf.
//
// gayconf is non-violent software: you can use, redistribute,
// and/or modify it under the terms of the CNPLv6+ as found
// in the LICENSE file in the source code root directory or
// at <https://git.pixie.town/thufie/CNPL>.
//
// gayconf comes with ABSOLUTELY NO WARRANTY, to the extent
// permitted by applicable law.  See the CNPL for details.
